from functools import wraps
import re
import flask
import jwt


AUTH_CONFIG_KEY = "authentication"


def authentication_enabled(auth_config):
    """Predicate on the auth configuration dictionary that returns whether
    authentication should be enabled for the application.

    :param auth_config: The authentication config dictionary.
    """
    if auth_config is None:
        return False
    return auth_config.get("enabled", True)


def authenticated(route):
    """Decorator for Flask routes that checks if the authentication header
    exists and is valid, and if it is, passes the user id as a named parameter
    `user_id` to the route.

    :param route: The route to wrap.
    """

    @wraps(route)
    def wrapper(*args, **kwargs):
        auth_config = flask.current_app.config[AUTH_CONFIG_KEY]
        if not authentication_enabled(auth_config):
            return route(*args, user_id=auth_config["user_id"], **kwargs)

        header = flask.request.headers.get("Authorization")
        if header is None:
            flask.abort(401, "Missing authorization header")

        try:
            claims = _extract_claims(auth_config, header)
        except ValueError as exc:
            flask.abort(401, str(exc))
        except (jwt.ExpiredSignatureError, jwt.InvalidTokenError):
            flask.abort(401, "Invalid or expired authentication token")

        user_id = claims.get("sub")
        if user_id is None:
            flask.abort(401, "Missing user id")
        return route(*args, user_id=user_id, **kwargs)

    return wrapper


_TOKEN_PATTERN = re.compile(r"Bearer (.*)")


def _extract_claims(auth_config, header):
    """Return a dictionary of claims extracted from the Authorization header.

    :param auth_config: The authentication config dictionary.
    :param header: The contents of the authorization header.
    """
    m = _TOKEN_PATTERN.match(header)
    if not m:
        raise ValueError("Invalid authorization header")

    token = m.group(1)
    return jwt.decode(
        token, auth_config.get("secret"), algorithms=["HS256"]
    )



"""
Derive information about the service from the environment.
"""
import logging
import subprocess
from functools import lru_cache

log = logging.getLogger(__name__)


@lru_cache(maxsize=1)
def read_version():
    """Attempts to read the version from the environment, returning 'NA' when it
    fails.  It first looks for the existence of a VERSION file, then tries to
    use git to obtain a tag.

    :returns: A string describing the version.
    """
    return _read_version_from_file() or _read_version_from_git() or "NA"


def _read_version_from_file():
    handle = None
    try:
        log.info("Reading version from file")
        handle = open("VERSION", "r")
        return handle.read().rstrip()
    except FileNotFoundError:
        return None
    finally:
        if handle:
            handle.close()


def _read_version_from_git():
    try:
        return (
            subprocess.run(["git", "describe", "--always"], stdout=subprocess.PIPE)
            .stdout.decode("utf-8")
            .rstrip()
        )
    except subprocess.SubprocessError:
        return None
    except FileNotFoundError:
        return None



"""
This file contains a hack to make swagger-ui work from behind a load balancer.
It contains no application specific code and could have been a library.
"""
from flask_restplus import Api
from flask_restplus.api import SwaggerView


class CustomAPI(Api):
    """
    Hack to serve swagger.json from the /api prefix.  This is a requirement for
    serving swagger-ui from behind a path-routing load balancer without
    additional proxies.
    """

    def _register_specs(self, app_or_blueprint):
        if self._add_specs:
            endpoint = str('specs')
            self._register_view(
                app_or_blueprint,
                SwaggerView,
                self.default_namespace,
                '{}/swagger.json'.format(self.prefix),
                endpoint=endpoint,
                resource_class_args=(self, ))
            self.endpoints.add(endpoint)



import logging
import os

from flask import Blueprint, Flask, url_for
from flask_restplus import Resource, apidoc

from apian.config import init_logging
from apian.meta import read_version
from apian.api import CustomAPI
from apian.auth import authentication_enabled, AUTH_CONFIG_KEY

log = logging.getLogger(__name__)


def create_api(name, config):
    # pylint: disable=unused-variable
    """Create an API with info and health endpoints.

    :param name: The name of the application; used in the generated frontend,
        and the URL prefix.
    :param config: The application configuration.

    :returns: An instance of :class:`~apian.api.CustomAPI`.
    """
    prefix = "/{}/api".format(name)
    title = _title_from_name(name)
    blueprint = Blueprint(prefix, __name__)
    extra = {}
    if authentication_enabled(config.get(AUTH_CONFIG_KEY)):
        extra["security"] = "Bearer Auth"

        extra["authorizations"] = {
            "Bearer Auth": {"type": "apiKey", "in": "header", "name": "Authorization"}
        }

    api = CustomAPI(blueprint, title=title, prefix=prefix, doc=prefix + "/", **extra)

    @api.route("/info")
    class Info(Resource):
        def get(self):
            # pylint: disable=no-self-use
            """Show information about the service"""
            return {"name": name, "version": read_version(), "config": config["__path"]}

    @api.route("/health")
    class Health(Resource):
        def get(self):
            # pylint: disable=no-self-use
            """Check the health of the service"""
            return {"status": "ok"}

    return api


def create_app(api, config):
    """Create a WSGI app.

    :param api: The API to serve, should be an instance of :class:`~CustomAPI`.
    :param config: Configuration for the app.  The following keys have an
        effect: 'debug', 'environment', '__path', 'log_config'.

    :returns: A WSGI app.
    """

    init_logging(config.get("log_config"))
    app = Flask(__name__)
    app.config.update(config)

    app.config["DEBUG"] = config.get("debug", False)
    app.config["ENV"] = config.get("environment", "production")

    swagger = _create_swagger()
    app.register_blueprint(api.blueprint)
    app.register_blueprint(swagger, url_prefix=api.prefix)

    return app


def _create_swagger():
    swagger = apidoc.Apidoc(
        "restplus_custom_doc",
        __name__,
        template_folder="templates",
        static_folder=os.path.dirname(apidoc.__file__) + "/static",
        static_url_path="/swagger",
    )

    @swagger.add_app_template_global
    def swagger_static(filename):
        # pylint: disable=unused-variable
        return url_for("restplus_custom_doc.static", filename=filename)

    return swagger


def _is_authenticated(config):
    if "authentication" not in config:
        return False
    return config["authentication"].get("enabled", True)


def _title_from_name(name):
    return " ".join(name.split("-")).title()
